#include "r2k.h"
#include "heap.h"

#define ALIGN_SIZE      4
#define ALIGN_MASK      (ALIGN_SIZE - 1)
#define ROUNDUP(size)   (((size_t)(size) + ALIGN_MASK) & ~ALIGN_MASK)

/* Initialise an empty heap. */

void HeapInit(Heap* heap, const HeapFunctions* fns)
{
	heap->free_list.next = &heap->free_list;
	heap->free_list.size = 0;
	heap->free_list.vm_size = 0;
	heap->scan_head = &heap->free_list;
	heap->fn = fns;
}

/* Add a new block of memory to the free list. */

struct heap_header* HeapExtend(Heap* heap, void* ptr, size_t size)
{
	struct heap_header* p = ptr;

	p->size = size;
	p->vm_size = size;

	/* Insert to free list */
	struct heap_header* prev;
	for (prev = heap->scan_head; !(p > prev && p < prev->next); prev
			= prev->next)
	{
		if (prev >= prev->next && (p > prev || p < prev->next))
			break;
	}
	p->next = prev->next;
	prev->next = p;
	heap->scan_head = prev;
	return prev;
}

static struct heap_header *more_core(Heap* heap, size_t size)
{
	void* p = heap->fn->claim(heap, &size);
	if (!p)
		return NULL;

	return HeapExtend(heap, p, size);
}

void* HeapAlloc(Heap* heap, size_t size)
{
	struct heap_header *p, *prev;

	if (size == 0) /* sanity check */
		return NULL;
	size = ROUNDUP(size + sizeof(struct heap_header));

	if (heap->scan_head == NULL)
	{
		/* Initialize */
		heap->free_list.next = &heap->free_list;
		heap->free_list.size = 0;
		heap->free_list.vm_size = 0;
		heap->scan_head = &heap->free_list;
	}
	prev = heap->scan_head;
	for (p = prev->next;; prev = p, p = p->next)
	{
		if (p->size >= size)
		{ /* big enough */
			if (p->size == size) /* exactly */
				prev->next = p->next;
			else
			{ /* allocate tail end */
				p->size -= size;
				p = (struct heap_header *) ((quad) p + p->size);
				p->size = size;
				p->vm_size = 0;
			}
#ifdef CONFIG_MCHECK
			p->magic = MALLOC_MAGIC;
#endif
			heap->scan_head = prev;
			break;
		}
		if (p == heap->scan_head)
		{
			p = more_core(heap, size);
			if (!p)
				break;
		}
	}

	if (!p)
		return NULL;

	return (void *) (p + 1);
}

void HeapFree(Heap* heap, void *addr)
{
	struct heap_header *p, *prev;

	if (addr == NULL)
		return;

	p = (struct heap_header *) addr - 1;
#ifdef CONFIG_MCHECK
	if (p->magic != MALLOC_MAGIC)
	sys_panic("free: invalid pointer");
	p->magic = 0;
#endif
	for (prev = heap->scan_head; !(p > prev && p < prev->next); prev
			= prev->next)
	{
		if (prev >= prev->next && (p > prev || p < prev->next))
			break;
	}
	if ((prev->next->vm_size == 0) && /* join to upper block */
			((quad) p + p->size == (quad) prev->next))
	{
		p->size += prev->next->size;
		p->next = prev->next->next;
	}
	else
	{
		p->next = prev->next;
	}
	if ((p->vm_size == 0) && /* join to lower block */
			((quad) prev + prev->size == (quad) p))
	{
		prev->size += p->size;
		prev->next = p->next;
	}
	else
	{
		prev->next = p;
	}

	/* Deallocate pool, if possible */
	if (p->size == p->vm_size)
	{
		if (heap->fn->release(heap, p, p->vm_size))
			prev->next = p->next;
	}

	heap->scan_head = prev;
}

#ifdef CONFIG_MSTAT
void
mstat(void)
{
	struct heap_header *p;

	printf("mstat: task=%x\n", task_self());
	for (p = free_list.next; p != &free_list; p = p->next)
	{
		printf("mstat: addr=%x size=%d next=%x\n", p, p->size, p->next);
	}
}
#endif
