#include "r2-asm.h"

.extern R2_swi_table
.extern R2_system_control_handlers

/* When an SWI occurs, the monitor will (eventually) call this routine.
 * The monitor will set up, and partially populate, a stack frame.
 *
 *     +28
 *     +24 SWI number
 *     +20 pc to return to
 *     +16 -- (lr)
 *     +12 sp when SWI occurred
 *     +8  -- (r12)
 *     +4  -- (r11)
 *   sp+0  -- (r10)
 *         
 * The idea is that the SWI handler will save registers here and then be able
 * to load them again and return in a single ldm instruction.
 *
 * This routine will dispatch the SWI, and then return directly to the caller.
 * (Without going through the monitor. Huge speed savings.)
 *
 * BIG WARNING: BE VERY CAREFUL CHANGING ANY OF THIS! In particular, be very
 * wary about doing anything with the stack. The SWI handler *must* be entered
 * with the stack pointer unchanged, because it may want to push r0-r9 to get
 * a contiguous run of registers on the stack (which then turn into a Registers
 * structure for use in C). 
 */

.text
proc SWIDispatcher
	stmia sp, {r10, r11, r12}
	str lr, [sp, #16]

	ldr r11, [sp, #24]            // r11 = SWI number
	bic r11, r11, #X
	cmp r11, #0x100
	bge not_a_kernel_swi          // special treatment for anything above 0xff

	ldr r12, = SWITable
	add r12, r12, r11, lsl #3     // r12 = address of dispatch structure
	ldr r10, [r12, #0]            // r10 = dispatch routine
	ldr r12, [r12, #4]            // r12 = parameter
	
	/* Right now:
	 *   r10 = dispatch routine
	 *   r11 = SWI number, without X bit
	 *   r12 = user field
	 * We're now correctly set up to call an OS SWI as registered with
	 * OS_ClaimSWI.
	 */
	
	adr lr, ExitFromSWI
	mov pc, r10
	
proc ExitFromSWI
	ldr r11, [sp, #24]
	tst r11, #X                   // check X bit, and return immediately if set
	ldmneia sp, {r10, r11, r12, sp, lr, pc}
	ldmvcia sp, {r10, r11, r12, sp, lr, pc} // if !X, return immediately if !V
	/* fall through to error code */
	
	
	
	/* Generates an error.
	 *
	 * On entry:
	 *   r0 points to error block
	 */ 
	
swi_error:
	/* FIXME: need to issue service call here */
	
	mov r12, #Vector_ErrorV
	bl VProxy                     // call error vector
	
	/* If we get here, the error has been correctly claimed, so simply
	 * clear V and return. */
	 
	mrs r12, CPSR
	bic r12, r12, #PSR_V
	msr CPSR_f, r12
	ldmia sp, {r10, r11, r12, sp, lr, pc}



/* Call a SWI that isn't in the kernel table. We call into C to do the hard
 * work of looking up the routine and r12 values, and then call it.
 * 
 * On entry:
 *   r0-r9 = user parameters
 *   r11 = SWI without X bit
 */

not_a_kernel_swi:
	cmp r11, #0x200
	blt WriteI                     // special treatment for 0x100-0x1FF

	stmfd sp!, {r0-r3}             // C corrupts r0-r3 (also r12)
	
	bic r0, r11, #63
	bl FindModuleBySWIBase
	and r11, r11, #63
	mov r12, r0
	
	ldmfd sp!, {r0-r3}
	
	adr lr, ExitFromSWI
	ldr pc, [r12, #4]



/* Very simple SWI that just generates an error.
 *
 * On entry:
 *   r0 = pointer to error block
 */
 
proc Impl_OS_GenerateError
	msr CPSR_f, #PSR_V
	mov pc, lr
	


/* SWI dispatch table.
 *
 * This is used when dispatching kernel SWIs. It consists of 0x100 fields,
 * each of which are:
 *
 *   +4 user ptr
 *   +0 routine
 *
 * A routine of 0 means it's unimplemented. The user ptr varies according
 * to what the routine is.
 *
 *  CProxy: r12 = routine of C function to call
 *  VProxy: r12 = vector number
 */

.data 4
proc SWITable 
	.word VProxy, Vector_WrchV // --- 0x00
	.word Impl_OS_WriteS, 0
	.word Impl_OS_Write0, 0
	.word Impl_OS_NewLine, 0
	.word VProxy, Vector_RdchV
	.word VProxy, Vector_CliV
	.word VProxy, Vector_ByteV
	.word VProxy, Vector_WordV
	.word VProxy, Vector_FileV // 0x08
	.word VProxy, Vector_ArgsV
	.word VProxy, Vector_BGetV
	.word VProxy, Vector_BPutV
	.word VProxy, Vector_GBPBV
	.word VProxy, Vector_FindV
	.word VProxy, Vector_ReadLineV
	.word CProxy, UnknownSWI
	.word CProxy, Impl_OS_GetEnv // --- 0x10
	.word Impl_OS_Exit, 0
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI // 0x18
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, Impl_OS_Heap
	.word CProxy, Impl_OS_Module
	.word CProxy, Impl_OS_Claim
	.word CProxy, UnknownSWI // --- 0x20
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI // 0x28
	.word CProxy, Impl_OS_FSControl
	.word CProxy, UnknownSWI
	.word Impl_OS_GenerateError, 0
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, Impl_OS_ServiceCall // --- 0x30
	.word CProxy, Impl_OS_ReadVduVariables
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI // 0x38
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word VProxy, Vector_ChangeEnvironmentV // --- 0x40
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word Impl_OS_WriteN, 0
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI // 0x48
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI // --- 0x50
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, Impl_OS_AddCallback
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, Impl_OS_ReadSysInfo // 0x58
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, Impl_OS_ReadDynamicArea
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI // --- 0x60
	.word CProxy, UnknownSWI
	.word CProxy, Impl_OS_ClaimSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI // 0x68
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, Impl_OS_SynchroniseCodeAreas
	.word CProxy, UnknownSWI
	.word CProxy, Impl_OS_AMBControl // --- 0x70
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI // 0x78
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI // --- 0x80
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI // 0x88
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI // --- 0x90
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI // 0x98
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI // --- 0xa0
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI // 0xa8
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI // --- 0xb0
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI // 0xb8
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI // --- 0xc0
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI // 0xc8
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI // --- 0xd0
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, Impl_OS_ConvertCardinal // FIXME: these are actually different
	.word CProxy, Impl_OS_ConvertCardinal
	.word CProxy, Impl_OS_ConvertCardinal
	.word CProxy, Impl_OS_ConvertCardinal // 0xd8
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI // --- 0xe0
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI // 0xe8
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI // --- 0xf0
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI // 0xf8
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
	.word CProxy, UnknownSWI
