#include "r2k.h"

#if 0
OS_AMBControl (SWI &70)
----------------------

Control of application memory blocks.

entry:

    R0 = reason code (bits 0-7) and flags (bits 8-31, reason code specific)
    Other registers depend upon reason code

exit:

    R0 preserved
    Other registers depend upon reason code

Notes on use:

 - AMBControl is initialised during kernel initialisation;

 - it is the clients responsibility to avoid degenerate mappings to application
   space (map out before allocate or map in);

 - an allocation of 0 pages is allowed, and creates a node which may later grow;

 - a shrink to 0 pages also frees the node (handle returned as 0);


OS_AMBControl 0
---------------

Allocate an AMB. The allocated pages are mapped in to application space.

entry:
     R0 = reason code and flags
          bits 0-7   0 (reason code)
          bits 8-31  reserved (must be 0)
     R1 = number of pages to allocate

exit:
     R1 = number of pages allocated
     R2 = abstract handle for AMB, or 0 if allocation failed


OS_AMBControl 1
---------------

Deallocate an AMB

entry:
     R0 = reason code and flags
          bits 0-7   1 (reason code)
          bits 8-31  reserved (must be 0)
     R2 = handle

exit:
     R0,R2 preserved


OS_AMBControl 2
---------------

Change number of pages allocated to an AMB (grow or shrink). Note that a shrink
to 0 pages will free the node (handle returned as 0).

entry:
     R0 = reason code and flags
          bits 0-7   2 (reason code)
          bits 8-31  reserved (must be 0)
     R1 = new number of pages required
     R2 = handle

exit:
     R1 = new number of pages achieved
     R3 = old number of pages


OS_AMBControl 3
---------------

Map (some) pages of an AMB in or out of application space.

entry:
     R0 = reason code and flags
          bits 0-7   3 (reason code)
          bit  8     partial mapping only, if set (specified via registers R3,R4)
          bits 9-31  reserved (must be 0)

     R1 = start (logical) address of new mapping;
          special values: 0 means map in to application start; -1 means map out
     R2 = handle
     R3 = offset (in pages) to start of mapping (if bit 8 of R0 set)
     R4 = number of pages to map (if bit 8 of R0 set)

exit:
     registers R0-R4 preserved


OS_AMBControl 4
---------------

Return information on an AMB.

entry:
     R0 = reason code and flags
          bits 0-7   4 (reason code)
          bits 8-31  reserved (must be 0)
     R2 = handle

exit:
     R1 = current mapping start address; -1 means mapped out
     R3 = currently allocated number of pages


Possible future work:

Take control of AMB-like part of Wimps ClaimFreeMemory, and hence know when
free pool pages are dead. The latter allows avoidance of cache cleaning/flushing
when mapping pages out of free pool.

Offer other facilities; eg. uncacheable app space on a per node basis, to support
running old, dynamic code.
#endif

static void mapamb(AMB* amb)
{
	for (int i = 0; i < amb->pages; i++)
	{
		SWI2i0o(R2K_MapPhysicalPage,
				amb->pagearray[i], /* physical address */
				amb->mappingaddress + (i * PAGESIZE)); /* virtual address */
	}
}

AMB* CreateAndMapAMB(quad pages)
{
	AMB* amb = malloc(sizeof(AMB));
	if (!amb)
		return NULL;

	amb->pages = 0;
	amb->mappingaddress = 0x8000;
	amb->pagearraysize = pages;
	amb->pagearray = malloc(pages * sizeof(quad));
	if (!amb->pagearray)
	{
		free(amb);
		return NULL;
	}

	/* Claim physical memory from the system. */

	for (int i = 0; i < pages; i++)
	{
		quad physical;
		SWI0i1o(R2K_AllocatePhysicalPage, physical);
		if (!physical)
			break;

		//printf("page %x/%x at %x\n", i, pages, physical);
		amb->pagearray[i] = physical;
		amb->pages++;
	}

	/* Make sure all other pages are properly nulled out. */

	for (int i = amb->pages; i < pages; i++)
		amb->pagearray[i] = 0;

	mapamb(amb);
	return amb;
}

SYSCALL(Impl_OS_AMBControl)
{
	switch (R0 & 0xFF)
	{
		default:
			printf("Unhandled OS_AMBControl opcode %x!\n", R0 & 0xFF);
			abort();
			break;
	}

	return 0;
}

