#include "globals.h"
#include "sleevecc/dis.h"

static const char* regnames[] =
{
	"r0", "r1", "r2", "r3", "r4", "r5", "r6", "r7",
	"r8", "r9", "r10", "r11", "r12", "sp", "lr", "pc"
};

static DisOptions options =
{
	disopt_CommaSpace,
	regnames
};

static arm_word breakpoints[16];
static arm_word run_until = 0;

static void disassemble(arm_word a)
{
	try
	{
		arm_word w = mm_load32(a);
		Instruction* insn = instr_disassemble(w, a, &options);

		printf("%08x : %08x\t%s\n",	a, w, insn->text);
	}
	catch (const mm_exc_dataabort& e)
	{
		printf("%08x : unreadable\n", a);
	}
}

static void cmd_help()
{
	printf("h             get this help\n");
	printf("p             print current CPU status\n");
	printf("q             quit\n");
	printf("r             run\n");
	printf("r <address>   run until\n");
	printf("b             list breakpoints\n");
	printf("b <address>   set breakpoint\n");
	printf("d <n>         delete breakpoint\n");
	printf("x <address>   examine 128 bytes of memory from <address>\n");
	printf("u <address>   disassemble 16 instructions from <address>\n");
	printf("All addresses are in hexadecimal. Breakpoint numbers are in decimal.\n");
}

static void cmd_status()
{
	arm_word a = context.r[15];
	printf("-> "); disassemble(a);
	printf("   "); disassemble(a+4);
	printf("   "); disassemble(a+8);
	printf("   "); disassemble(a+12);

	for (int i=0; i<16; i++)
	{
		printf("% 3s:%08x ", regnames[i], context.r[i]);
		if ((i == 3) || (i == 7) || (i == 11))
			printf("\n");
	}

	printf(" %c%c%c%c\n",
			context.n ? 'N' : '-',
			context.z ? 'Z' : '-',
			context.c ? 'C' : '-',
			context.v ? 'V' : '-');

	printf("stack: ");
	for (int i=0; i<8; i++)
	{
		try
		{
			arm_word s = mm_load32(context.r[13] + i*4);
			printf("%08x ", s);
		}
		catch (const mm_exc_dataabort& e)
		{
			printf("xxxxxxxx ");
		}
	}
	printf("\n");
}

static void cmd_run()
{
	int count = 10000000;

	while (context.r[15] != run_until)
	{
		int i = singlestep();
		if (i)
			break;

		count--;
		if (count == 0)
		{
			printf("Watchdog break\n");
			return;
		}

		for (int i = 0; i < sizeof(breakpoints)/sizeof(*breakpoints); i++)
			if (context.r[15] == breakpoints[i])
			{
				printf("Stopped on breakpoint %d\n", i);
				return;
			}
	}
}

static void cmd_breakpoint_list()
{
	printf("Breakpoints:\n");

	for (int i = 0; i < sizeof(breakpoints)/sizeof(*breakpoints); i++)
	{
		if (breakpoints[i])
		{
			printf("% 3d : %08x\n",
					i, breakpoints[i]);
		}
	}
}

static void cmd_breakpoint_set(arm_word address)
{
	for (int i = 0; i < sizeof(breakpoints)/sizeof(*breakpoints); i++)
	{
		if (!breakpoints[i])
		{
			breakpoints[i] = address;
			printf("Set breakpoint %d to %08x\n", i, address);
			return;
		}
	}

	printf("No breakpoint slots left!\n");
}

static void cmd_breakpoint_delete(int i)
{
	if ((i < 0) || (i >= sizeof(breakpoints)/sizeof(*breakpoints)))
	{
		printf("Breakpoint out of range!\n");
		return;
	}

	breakpoints[i] = 0;
	printf("Breakpoint removed.\n");
}

static void cmd_examine(arm_word address)
{
	for (arm_word line = address; line < (address + 128); line += 16)
	{
		printf("%08x : ", line);

		for (int i=0; i<16; i++)
		{
			try
			{
				printf("%02x ", mm_load8(line + i));
			}
			catch (const mm_exc_dataabort& e)
			{
				printf("xx ");
			}
		}

		printf(": ");

		for (int i=0; i<16; i++)
		{
			try
			{
				int c = mm_load8(line + i);
				if ((c < 32) || (c > 126))
					c = '.';
				printf("%c", c);
			}
			catch (const mm_exc_dataabort& e)
			{
				printf(".");
			}
		}

		printf("\n");
	}
}

static void cmd_unassemble(arm_word address)
{
	for (arm_word offset = 0; offset < (16*4); offset += 4)
		disassemble(address + offset);
}

void debugloop()
{
	char buffer[80];

	cmd_status();

	for (;;)
	{
		printf("\n@ ");
		fflush(stdout);

		fgets(buffer, sizeof(buffer), stdin);

		switch (buffer[0])
		{
			case 'h':
				cmd_help();
				break;

			case 'p':
				cmd_status();
				break;

			case 'q':
				exit(0);

			case 'n':
			case '\0':
			{
				int s = singlestep();
				cmd_status();
				break;
			}

			case 'r':
			{
				int i = sscanf(buffer, "r %x", &run_until);

				if (i != 1)
					run_until = 0;

				cmd_run();
				printf("\n");
				cmd_status();
				break;
			}

			case 'b':
			{
				arm_word address = 0;

				int i = sscanf(buffer, "b %x", &address);
				if (i > 0)
					cmd_breakpoint_set(address);
				else
					cmd_breakpoint_list();
				break;
			}

			case 'd':
			{
				int n;

				int i = sscanf(buffer, "d %d", &n);
				if (i > 0)
					cmd_breakpoint_delete(n);
				else
					printf("You must specify a breakpoint number.\n");

				break;
			}

			case 'x':
			{
				arm_word address;

				int i = sscanf(buffer, "x %x", &address);
				if (i > 0)
					cmd_examine(address);
				else
					printf("You must specify a start address.\n");
				break;
			}

			case 'u':
			{
				arm_word address;

				int i = sscanf(buffer, "u %x", &address);
				if (i > 0)
					cmd_unassemble(address);
				else
					printf("You must specify a start address.\n");
				break;
			}

			default:
				printf("Unknown command --- try h for help\n");
		}
	}
}
